print("import: 'pre_commit'")
import pre_commit

print("import: 'pre_commit.commands'")
import pre_commit.commands

print("import: 'pre_commit.languages'")
import pre_commit.languages

print("import: 'pre_commit.meta_hooks'")
import pre_commit.meta_hooks

print("import: 'pre_commit.resources'")
import pre_commit.resources

